import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import javax.imageio.ImageIO;

import org.neuroph.core.NeuralNetwork;
import org.neuroph.core.learning.SupervisedTrainingElement;
import org.neuroph.core.learning.TrainingSet;
import org.neuroph.nnet.MultiLayerPerceptron;
import org.neuroph.util.TransferFunctionType;


public class Run {
public static final int INPUT_VECTOR_SIZE = 20*20*3;
public static final int OUTPUT_VECTOR_SIZE = 3;

   public static void testNeural() {
	   
	   NeuralNetwork nn = NeuralNetwork.load("networks/latest.nnet");
	   
	   nn.setInput(getTrainingSet((new File("imgs/sqsq.png"))).getInput());
	   nn.calculate();
	   double[] output = nn.getOutput();
	   for (double dd : output) {
		   System.out.println(dd);
	   }
   }

	public static void main(String[] args) {
		//testNeural();
		//if (true) return;
		
			ArrayList<File> images = listFilesForFolder(new File("./imgs/"));
		    
			
			TrainingSet<SupervisedTrainingElement> trainings = new TrainingSet<SupervisedTrainingElement>(INPUT_VECTOR_SIZE, OUTPUT_VECTOR_SIZE);
			
			
			for (File f :images) {
				trainings.addElement(getTrainingSet(f));
			}
			
 	     	MultiLayerPerceptron mlp = new MultiLayerPerceptron(TransferFunctionType.SIGMOID, INPUT_VECTOR_SIZE,  24, 12, OUTPUT_VECTOR_SIZE);
 	     	
 	     	System.out.println("training");
		    mlp.learn(trainings);

		    mlp.save("./networks/latest.nnet");
		    
		    System.out.println("done");
		    
	}
	
	public static SupervisedTrainingElement getTrainingSet(File file) {
		ArrayList<Double> input = null;
		ArrayList<Double> output = new ArrayList<>();
		
		//output
		
		if (file.getName().startsWith("circle")) {
			output.add(1.0d);
			output.add(0.0d);
			output.add(0.0d);
			System.out.println("Found circle: " + file.getName());
		} else if (file.getName().startsWith("other")) {
			output.add(0.0d);
			output.add(1.0d);
			output.add(0.0d);
			System.out.println("Found other: " + file.getName());
		}  else if (file.getName().startsWith("square")) {
			output.add(0.0d);
			output.add(0.0d);
			output.add(1.0d);	
			System.out.println("Found square: " + file.getName());
		} 
		
		// input
		try {
	     	input = getInput(file);
	     //	System.out.println(raster.getWidth() + ", " + raster.getHeight() + " : " +  INPUT_VECTOR_SIZE + "vs" + input.size());
		} catch (Exception e ) {
			e.printStackTrace();
		}
		
		return new SupervisedTrainingElement(input, output);
	}
	
	public static ArrayList<Double> getInput(File file) throws IOException {
		ArrayList<Double> input = new ArrayList<>();
		
		
		final BufferedImage bf = ImageIO.read(file);
     	final Raster raster = bf.getData();
     	
     	for (int i = 0 ; i < raster.getHeight(); i++) {
     		for (int j = 0; j < raster.getWidth(); j++) {
     			double[] buff = new double[3];
     		    raster.getPixel(i, j, buff);
     			
     			input.add(buff[0] / 255.0d);
     			input.add(buff[1] / 255.0d);
     			input.add(buff[2] / 255.0d);
     		}
     	}
     	
     	return input;
	}
	
	
	public static ArrayList<File> listFilesForFolder(final File folder) {
	    ArrayList<File> files = new ArrayList<File>();
		
		for (final File fileEntry : folder.listFiles()) {
	        if (fileEntry.isDirectory()) {
	        } else {
	            files.add(fileEntry);
	        }
	    }
		
		return files;
	}

}
