package soft.robot;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.Timer;
import java.util.TimerTask;

import org.neuroph.contrib.imgrec.ImageRecognitionPlugin;
import org.neuroph.contrib.imgrec.ImageSizeMismatchException;
import org.neuroph.contrib.imgrec.image.Color;
import org.neuroph.core.NeuralNetwork;

import android.app.Activity;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.ImageFormat;
import android.graphics.Rect;
import android.graphics.YuvImage;
import android.hardware.Camera;
import android.hardware.Camera.Size;
import android.os.Bundle;
import android.os.Environment;
import android.util.Log;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.View;
import android.view.Window;
import android.view.WindowManager;
import android.widget.Button;
import android.widget.TextView;
import at.abraxas.amarino.Amarino;
import at.abraxas.amarino.AmarinoIntent;

public class MainActivity extends Activity {

	private static final String TAG = "MainActivity";
	SurfaceView mVideoCaptureView;
    private Camera mCamera;
    
    private final String CURRENT_SEE_FILENAME = "see.png";
    private static final int DETECT_IMAGE_PERIOD = 1000;
    private static final int SCALE_WIDTH = 20;
    private static final int SCALE_HEIGHT = 20;
    private static final double REQUIRED_BELIEF = 0.7;
    private static final int NETWORK_ID = R.raw.latest;
    
    private Button btnSquare;
    private Button btnCircle;
    private Button btnOther;
    private TextView txtBelief;
    
	private NeuralNetwork nnet;
	private ImageRecognitionPlugin imageRecognition;
	
	private Belief belief = new Belief();
	private Timer detectionScheduler = new Timer();
    
	private static final int THREAD_STACK_SIZE = 120536;
	
	private static final String DEVICE_ADDRESS =  "00:12:02:10:41:02";
	private ArduinoReceiver arduinoReceiver = new ArduinoReceiver();
    
	
	@Override
	protected void onStart() {
		super.onStart();
		registerReceiver(arduinoReceiver, new IntentFilter(AmarinoIntent.ACTION_RECEIVED));
		
		Amarino.connect(MainActivity.this, DEVICE_ADDRESS);
				
	}


	@Override
	protected void onStop() {
		super.onStop();
		Amarino.disconnect(this, DEVICE_ADDRESS);
		unregisterReceiver(arduinoReceiver);
	}
	
	public class ArduinoReceiver extends BroadcastReceiver {

		@Override
		public void onReceive(Context context, Intent intent) {
			String data = null;
			final String address = intent.getStringExtra(AmarinoIntent.EXTRA_DEVICE_ADDRESS);
			final int dataType = intent.getIntExtra(AmarinoIntent.EXTRA_DATA_TYPE, -1);
			
			if (dataType == AmarinoIntent.STRING_EXTRA){
				data = intent.getStringExtra(AmarinoIntent.EXTRA_DATA);
				
				if (data != null){
					try {
						
						// ak v buducnosti chceme pridat komunikaciu z druhej strany
					} 
					catch (NumberFormatException e) { 
						
					}
				}
			}
		}
	}
	
	private void decide() {
		if (belief.getShape() != Shape.NULL) {
			if (belief.confidence >= REQUIRED_BELIEF) {
				
				switch (belief.getShape()) {
				case CIRCLE :
					Log.v(TAG, "I see circle!");
					Amarino.sendDataToArduino(MainActivity.this, DEVICE_ADDRESS, 'c', belief.getConfidence());
					
					break;
					
				case SQUARE:
					Log.v(TAG, "I see square!");
					Amarino.sendDataToArduino(MainActivity.this, DEVICE_ADDRESS, 's', belief.getConfidence());
					
					break;
				}
			}
		}
		else {
			Log.v(TAG, "Nothing!");
			Amarino.sendDataToArduino(MainActivity.this, DEVICE_ADDRESS, 'n', belief.getConfidence());
			
		}
	}
	
    private final String getCurrentSeeFile() {
    	return MainActivity.this.getExternalFilesDir(Environment.DIRECTORY_DOWNLOADS)+"//"+CURRENT_SEE_FILENAME;
    }
    
    private final String getFilePath(String filename) {
    	return MainActivity.this.getExternalFilesDir(Environment.DIRECTORY_DOWNLOADS)+"//"+filename;
    }
    
    private SurfaceHolder.Callback2 mSurfaceCallback = new SurfaceHolder.Callback2() {
		
		@Override
		public void surfaceDestroyed(SurfaceHolder holder) {
			
		}
		
		@Override
		public void surfaceCreated(SurfaceHolder holder) {
			cameraStart();
		}
		
		@Override
		public void surfaceChanged(SurfaceHolder holder, int format, int width,
				int height) {
			
		}
		
		@Override
		public void surfaceRedrawNeeded(SurfaceHolder holder) {
			
		}
	};
	
	
	private void detect() {
		   if (nnet == null)
			   return;
		   
		   synchronized(MainActivity.this) {
			Log.i(TAG, "Detection run! " + (imageRecognition == null) );
				try {
					cloneSee("see_detection.png");
					
					
					BitmapFactory.Options opt = new BitmapFactory.Options();
					opt.inPreferredConfig= Bitmap.Config.ARGB_8888;
					FileInputStream fis = new FileInputStream(getFilePath("see_detection.png"));
					
					Bitmap bmp=BitmapFactory.decodeStream(fis,null,opt);
					int width=bmp.getWidth();
					int height=bmp.getHeight();
					double[] input = new double[width*height*3];
					
					int px;
					int pos = 0;
					for (int i = 0; i < height; i++) {
					    for (int j = 0; j < width; j++)
					    {
					          px= bmp.getPixel(j,i);
					          input[pos] = Color.getRed(px);
					          input[pos+1] = Color.getGreen(px);
					          input[pos+2] = Color.getBlue(px);
					        		  
					          pos+=3;
					    }
					}
					
					
					nnet.setInput(input);
					nnet.calculate();
					double[] output = nnet.getOutput();
					for (double dd : output) {
						Log.v("OUTPUT", dd+"");
					}
					Log.v("OUTPUT", "finished");
					belief.update(output);
				
					
				} catch (ImageSizeMismatchException e) {
					Log.e(TAG, "Image size mismatch! "+ e.getMessage());
					e.printStackTrace();
				} catch (Exception e) {
					Log.e(TAG, "Exception! "+ e.getMessage());
					e.printStackTrace();
				}
		   }
		   
		   decide();
	}
	
	private TimerTask detectionTask = new TimerTask() {
		
		@Override
		public void run() {
			   detect();
		}
	};
	
	private Runnable loadDataRunnable = new Runnable() {
		public void run() {
			//nnet = NeuralNetwork.load(getFilePath(CURRENT_NET));
			nnet = NeuralNetwork.load(  getResources().openRawResource(NETWORK_ID)  );
			imageRecognition = (ImageRecognitionPlugin) nnet.getPlugin(ImageRecognitionPlugin.class);
			
			Log.i(TAG, "Neural network loaded");
	        detectionScheduler.schedule(detectionTask, 0, DETECT_IMAGE_PERIOD);
		}
	};
    
	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		
		// requestWindowFeature(Window.FEATURE_NO_TITLE);
		// getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN,WindowManager.LayoutParams.FLAG_FULLSCREEN);
		// 
		

		loadNetwork();

		
		setContentView(R.layout.activity_main);
		
		
		btnSquare = (Button)findViewById(R.id.btnSquare);
		btnOther = (Button)findViewById(R.id.btnOther);
		btnCircle = (Button)findViewById(R.id.btnCircle);
		txtBelief = (TextView)findViewById(R.id.textBelief);
		
		setButtonFunctionality();
		
	
		 mVideoCaptureView = (SurfaceView) findViewById(R.id.video_capture_surface);
	        SurfaceHolder videoCaptureViewHolder = mVideoCaptureView.getHolder();
	       // videoCaptureViewHolder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS);
	        
	        videoCaptureViewHolder.addCallback(mSurfaceCallback);
	        
	     mVideoCaptureView.setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View v) {
				detect();
			}
		})   ;
	        
	     
	        
	}
	
	private void loadNetwork() {
		new Thread(null, loadDataRunnable, "dataLoader", THREAD_STACK_SIZE).start();
	}
	
	private void setButtonFunctionality() {
		btnSquare.setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View v) {
				cloneSee("square-"+System.currentTimeMillis()+".png");
			}
		});
		
		btnCircle.setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View v) {
				cloneSee("circle-"+System.currentTimeMillis()+".png");
			}
		});
		
		btnOther.setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View v) {
				cloneSee("other-"+System.currentTimeMillis()+".png");
			}
		});
		
	}
	
	private void cloneSee(String newFilename) {
		
		InputStream inStream = null;
		OutputStream outStream = null;
		
		try{
			 
    	    File afile =new File(getCurrentSeeFile());
    	    File bfile =new File(getFilePath(newFilename));
 
    	    inStream = new FileInputStream(afile);
    	    outStream = new FileOutputStream(bfile);
 
    	    byte[] buffer = new byte[1024];
 
    	    int length;
    	    
    	    while ((length = inStream.read(buffer)) > 0){
    	    	outStream.write(buffer, 0, length);
    	    }
 
    	    inStream.close();
    	    outStream.close();
 
    	}catch(IOException e){
    		e.printStackTrace();
    	}
    
	}

	public void cameraStart() {
		 SurfaceHolder videoCaptureViewHolder = null;
	        try {
	            mCamera = Camera.open();
	        } catch (RuntimeException e) {
	            Log.e(TAG, "Camera Open filed");
	            return;
	        }
	        
	        final Camera.Parameters parameters = mCamera.getParameters();
	        parameters.setPreviewFpsRange(15000,30000);
	        List<int[]> supportedPreviewFps=parameters.getSupportedPreviewFpsRange();
	        Iterator<int[]> supportedPreviewFpsIterator=supportedPreviewFps.iterator();
	        
	        while(supportedPreviewFpsIterator.hasNext()){
	            int[] tmpRate=supportedPreviewFpsIterator.next();
	            StringBuffer sb=new StringBuffer();
	            sb.append("supportedPreviewRate: ");
	            for(int i=tmpRate.length,j=0;j<i;j++){
	                sb.append(tmpRate[j]+", ");
	            }
	            Log.v(TAG,sb.toString());
	        }
	        
	        List<Size> supportedPreviewSizes=parameters.getSupportedPreviewSizes();
	        Iterator<Size> supportedPreviewSizesIterator=supportedPreviewSizes.iterator();
	        while(supportedPreviewSizesIterator.hasNext()){
	            Size tmpSize=supportedPreviewSizesIterator.next();
	            Log.v(TAG,"supportedPreviewSize.width = "+tmpSize.width+"supportedPreviewSize.height = "+tmpSize.height);
	        }
	        
	        mCamera.setDisplayOrientation(90);
	        
	        mCamera.setParameters(parameters);
	        if (null != mVideoCaptureView)
	            videoCaptureViewHolder = mVideoCaptureView.getHolder();
	        try {
	            mCamera.setPreviewDisplay(videoCaptureViewHolder);
	            
	        } catch (Throwable t) {
	        	Log.e(TAG, t.getMessage());
	        } 
	        
	        
	        
	        final Size previewSize=mCamera.getParameters().getPreviewSize();
	        
	        int dataBufferSize=(int)(previewSize.height*previewSize.width*
	                               (ImageFormat.getBitsPerPixel(mCamera.getParameters().getPreviewFormat())/8.0));

	        mCamera.addCallbackBuffer(new byte[dataBufferSize]);
	        mCamera.addCallbackBuffer(new byte[dataBufferSize]);
	        mCamera.addCallbackBuffer(new byte[dataBufferSize]);
	        
	        final Rect rect = new Rect(0, 0, previewSize.width, previewSize.height);
	        mCamera.setPreviewCallbackWithBuffer(new Camera.PreviewCallback() {
	            private long timestamp=0;
	            private FileOutputStream out;
	            private Bitmap bmp;
	            final String seeFile = MainActivity.this.getCurrentSeeFile();
	            
	            public synchronized void onPreviewFrame(byte[] data, Camera camera) {
	            	
	            	synchronized (MainActivity.this) {
	               // Log.v(TAG,"Time Gap = "+(System.currentTimeMillis()-timestamp));
		                timestamp=System.currentTimeMillis();
		              try{
		                    try {
		                    
		                    	YuvImage image = new YuvImage(data, parameters.getPreviewFormat(), previewSize.width, previewSize.height, null);
		                        
		                    	OutputStream outStream = null;
		                         try 
		                         {
			                             outStream = new FileOutputStream( seeFile );
			                             image.compressToJpeg(rect, 100, outStream);
			                             outStream.flush();
			                             outStream.close();
			                            
			                             bmp = BitmapFactory.decodeFile( seeFile );
			                             bmp = Bitmap.createScaledBitmap(bmp, SCALE_WIDTH, SCALE_HEIGHT, false);
			                             
			                             out = new FileOutputStream( seeFile );
			                             bmp.compress(Bitmap.CompressFormat.PNG, 100, out);
			                             out.flush();
			                             out.close();
		                             
		                         } 
		                         catch (FileNotFoundException e) 
		                         {
		                             e.printStackTrace();
		                         }
		                         catch (IOException e) 
		                         {
		                             e.printStackTrace();
		                         }
		                         
		                    } catch (Exception e1) {
		                        e1.printStackTrace();
		                    }           
		                    camera.addCallbackBuffer(data);
		                }catch (Exception e) {
		                    Log.e("CameraTest", "addCallbackBuffer error");
		                    return;
		                }
		                return;
		            }
	            }
	        });
	        try {
	            mCamera.startPreview();
	        } catch (Throwable e) {
	            mCamera.release();
	            mCamera = null;
	            return;
	        }
	}
	
	private void stopVideo() {
        if(null==mCamera)
            return;
        try {
            mCamera.stopPreview();
            mCamera.setPreviewDisplay(null);
            mCamera.setPreviewCallbackWithBuffer(null);
            mCamera.release();
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        mCamera = null;
    }
    public void finish(){
        stopVideo();
        super.finish();
    };

}
